#!/bin/bash
set -e
set -u

# Location of CBMC binary
CBMC=./src/cbmc/cbmc

# Location of EdgeMaker binary
MAKER=./HBFixer/install/bin/EdgeMaker

# Location of EdgeFixer binary
FIXER=./HBFixer/install/bin/EdgeFixer

# Location of HBFixer pre-processor
PREPROC=./HBFixer/EdgeMaker/pre_process.sh

die () {
  if [ ! -z "${@:-}" ]
  then
    echo ${@}
  fi
  exit 1
}

if [ -z ${1:-} ]
then
  die "First argument must be program to test"
fi

if [ ! -f ${1:-} ]
then
  die "Program not found: $1"
fi

if [ ! -f $CBMC ]
then
  die "CBMC binary not found: $CBMC"
fi

if [ ! -f $MAKER ]
then
  die "EdgeMaker binary not found: $MAKER"
fi

if [ ! -f $FIXER ]
then
  die "EdgeFixer binary not found: $FIXER"
fi

if [ ! -f $PREPROC ]
then
  die "Pre-processor script not found: $PREPROC"
fi

# Temporary files for various programs
CBMC_OUT=/tmp/cbmc.out
FIX_OUT=/tmp/fixer.out

# First, run CBMC
$CBMC $1 >$CBMC_OUT 2>/dev/null || true
$PREPROC ${CBMC_OUT}
$MAKER >$FIX_OUT
$FIXER $1 $FIX_OUT
