-- 
-- Author: Markus Kusano
--
-- Parser for formula file. 
--
-- The formula is a 

module FormulaParser (parseFormulaFile, VarId, Formula) where
import Text.ParserCombinators.Parsec
import Control.Applicative ((<$>), (*>), (<*>))

-- Take a list of each line in a file and parse it into a list of formulas
parseFormulaFile :: [String] -> [Either ParseError Formula]
--parseFormulasFile s = rights $ map (parse line "(error)") s
parseFormulaFile s = map (parse line "(error)") s

-- A variable ID is an integer
type VarId = Int

-- A formula is a list of variables
type Formula = [VarId]

-- A formula is a bunch of VarIDs (integers) separated by spaces
line :: GenParser Char st Formula
line = do
          spaces
          e1 <- integer
          others <- remainingInts
          return (e1 : others)

-- Recursively calls line if there is more input
remainingInts :: GenParser Char st Formula
--remainingEdges = spaces >> ((char ' ' >> line) <|> (return []))
remainingInts = ((char ' ' >> line) <|> (return []))

-- Parse an integer
integer :: GenParser Char st Int
integer = rd <$> (plus <|> minus <|> number)
    where rd     = read :: String -> Int
          plus   = char '+' *> number
          minus  = (:) <$> char '-' <*> number
          number = many1 digit
