/*******************************************************************\

Module: Symbolic Execution of ANSI-C

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <fstream>
#include <cstdlib>
#include <iostream>
#include <memory>
#include <algorithm>
#include <unistd.h>

#include <util/i2string.h>
#include <util/location.h>
#include <util/time_stopping.h>
#include <util/message_stream.h>

#include <langapi/mode.h>
#include <langapi/languages.h>
#include <langapi/language_util.h>

#include <ansi-c/ansi_c_language.h>

#include <goto-programs/xml_goto_trace.h>

#include <goto-symex/build_goto_trace.h>
#include <goto-symex/slice.h>
#include <goto-symex/slice_by_trace.h>
#include <goto-symex/memory_model_sc.h>
#include <goto-symex/memory_model_tso.h>
#include <goto-symex/memory_model_pso.h>

#include "bmc.h"
#include "bv_cbmc.h"

#include "compute_unsat_core.h"
#include "msun.h"

/*******************************************************************\

Function: bmct::do_unwind_module

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::do_unwind_module(
  decision_proceduret &decision_procedure)
{
}

/*******************************************************************\

Function: bmct::error_trace

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::error_trace( prop_convt &prop_conv)
{
   status() << "Building error trace" << eom;
   
  goto_tracet goto_trace;
  build_goto_trace(equation, prop_conv, ns, goto_trace, input_arg_properties);
  
  #if 0
  if(options.get_option("vcd")!="")
  {
    if(options.get_option("vcd")=="-")
      output_vcd(ns, goto_trace, std::cout);
    else
    {
      std::ofstream out(options.get_option("vcd").c_str());
      output_vcd(ns, goto_trace, out);
    }
  }
  #endif
  
  switch(ui)
  {
  case ui_message_handlert::PLAIN:
     std::cout << "\n" << "Counterexample:" << "\n";
     show_goto_trace(std::cout, ns, goto_trace);
     /*SEPI*/
     std::cout<<"------------- VT BugAssist ;) ------------- : "<<std::endl;
     unsat_core(prop_conv);
     /*SEPI*/
    break;
  
  case ui_message_handlert::XML_UI:
    {
      xmlt xml;
      convert(ns, goto_trace, xml);
      std::cout << xml << "\n";
    }
    break;
  
  default:
    assert(false);
  }
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
bool copy_file(const char* From, const char* To, std::size_t MaxBufferSize = 1048576)
{
    std::ifstream is(From, std::ios_base::binary);
    std::ofstream os(To, std::ios_base::binary);

    std::pair<char*,std::ptrdiff_t> buffer;
    buffer = std::get_temporary_buffer<char>(MaxBufferSize);

    //Note that exception() == 0 in both file streams,
    //so you will not have a memory leak in case of fail.
    while(is.good() and os)
    {
       is.read(buffer.first, buffer.second);
       os.write(buffer.first, is.gcount());
    }

    std::return_temporary_buffer(buffer.first);

    if(os.fail()) return false;
    if(is.eof()) return true;
    return false;
}
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*******************************************************************\

Function: bmct::write_check_file

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
#if 1
void bmct::write_check_file(int numOfOrigClauses, std::string new_eqn, std::string prev_eqn, std::vector<std::string> new_unsat_core, 
        std::vector<int> assertions)
{
    std::ofstream newEqn;
    newEqn.open(new_eqn.data());
    
    std::ifstream prevEqn(prev_eqn.data());

    std::string line;
    std::vector<std::string> assrt_split;

    std::getline(prevEqn, line); //getting the first line to get some information
    std::vector<std::string> first_line = split(std::string(line), ' ');
    std::string numOfVars = first_line.at(2);
    std::string numOfClauses = intToString(numOfOrigClauses);
    std::string weight = first_line.at(4);

    
    newEqn << "p wcnf " << numOfVars << " " << atoi(numOfClauses.c_str())-all_assertions.size()-input_arg_properties.size()+added_clauses.size()+2 << " " << weight;
    while (numOfOrigClauses > 0){
        std::getline(prevEqn, line);
        assrt_split = split(line, ' ');
        if(assrt_split.size() == 3){    //to avoid changing the lines with more than one variable (those whose first variable is the assertion variable)
            int val = atoi((assrt_split.at(1)).c_str());
            if(std::find(assertions.begin(), assertions.end(), val) != assertions.end()){
                //do nothing. We will add the disjunction of negation of assertions later in this function
            }
            else if(std::find(input_arg_properties.begin(), input_arg_properties.end(), val) != input_arg_properties.end()){
                //do nothing. We do not need the input argument constraints in our check_file#.cnf s
            }
            else if(std::find(input_arg_properties.begin(), input_arg_properties.end(), -val) != input_arg_properties.end()){
                //do nothing.
            }
            else{
                newEqn << "\n" << line;
            }
        }
        else{
            newEqn << "\n" << line;
        }
        numOfOrigClauses--;
    }
    for(int i = 0 ; i < added_clauses.size() ; i++)
        newEqn << "\n" << added_clauses.at(i);
    
    /*for(int i = 0 ; i < new_unsat_core.size() ; i++)
        std::cout << "$$ " << new_unsat_core.at(i) << "\n";*/
    //Adding the disjunction of negation of newly found UNSAT core clauses
    line = weight + negate_disjunction(new_unsat_core) + " 0";    //new added clause
    newEqn << "\n" << line;
    std::cout << "negate disjunction : " << line << "\n";
    added_clauses.push_back(line);  //This vector is what I will return at the end of the verification (or not exactly this?)

    //Finally, adding the disjunction of negation of assertions line:
    line = weight + negate_disjunction_no_weight(all_assertions) + " 0";
    newEqn << "\n" << line;
    
    newEqn.close();
    prevEqn.close();
}
#else
void bmct::write_check_file(int numOfOrigClauses, std::string new_eqn, std::string prev_eqn, std::vector<std::string> new_unsat_core, 
        std::vector<int> assertions)
{
    std::ofstream newEqn;
    newEqn.open(new_eqn.data());
    
    std::ifstream prevEqn(prev_eqn.data());

    std::string line;
    std::vector<std::string> assrt_split;

    std::getline(prevEqn, line); //getting the first line to get some information
    std::vector<std::string> first_line = split(std::string(line), ' ');
    std::string numOfVars = first_line.at(2);
    std::string numOfClauses = intToString(numOfOrigClauses);
    std::string weight = first_line.at(4);

    
    newEqn << "p cnf " << numOfVars << " " << atoi(numOfClauses.c_str())-all_assertions.size()+added_clauses.size()+2;// << " " << weight;
    while (numOfOrigClauses > 0){
        std::getline(prevEqn, line);
        assrt_split = split(line, ' ');
        if(assrt_split.size() == 3){    //to avoid changing the lines with more than one variable (those whose first variable is the assertion variable)
            int val = atoi((assrt_split.at(1)).c_str());
            if(std::find(assertions.begin(), assertions.end(), val) != assertions.end()){
                //do nothing. We will add the disjunction of negation of assertions later in this function
            }
            else{
	      newEqn << "\n" << assrt_split.at(1) << " " << assrt_split.at(2);
            }
        }
        else{
	  newEqn << "\n";
	  for (int i=1; i<assrt_split.size(); i++) {
	    newEqn << assrt_split.at(i);
	    if (i != assrt_split.size()-1)
	      newEqn << " ";
	  }
	}
        numOfOrigClauses--;
    }
    for(int i = 0 ; i < added_clauses.size() ; i++) {
      assrt_split = split(added_clauses.at(i), ' ');
      newEqn << "\n";
      for (int i=1; i<assrt_split.size(); i++) {
	newEqn << assrt_split.at(i);
	if (i != assrt_split.size()-1)
	  newEqn << " ";
      }
    }

    /*for(int i = 0 ; i < new_unsat_core.size() ; i++)
        std::cout << "$$ " << new_unsat_core.at(i) << "\n";*/
    //Adding the disjunction of negation of newly found UNSAT core clauses
    line = weight + negate_disjunction(new_unsat_core) + " 0";    //new added clause
    std::string myline = negate_disjunction(new_unsat_core) + " 0" ;
      newEqn << "\n" << (myline.c_str() + 1);
    std::cout << "negate disjunction : " << line << "\n";
    added_clauses.push_back(line);  //This vector is what I will return at the end of the verification (or not exactly this?)

    //Finally, adding the disjunction of negation of assertions line:
    line = negate_disjunction_no_weight(all_assertions) + " 0";
    newEqn << "\n" << (line.c_str()+1);
    
    newEqn.close();
    prevEqn.close();
}
#endif
/*******************************************************************\

Function: bmct::write_new_unsat_file

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
void bmct::write_new_unsat_file(prop_convt &prop_conv, std::string new_eqn, std::string prev_eqn, std::string ce_assignment, 
        std::vector<std::string> new_unsat_core){
    int numOfOrigClauses = prop_conv.get_clause_count();
    int hb_var_start = prop_conv.get_hb_var_start();
    //first, writing the original part of the cnf formula, the clauses
    std::ofstream newEqn;
    newEqn.open(new_eqn.data());
    
    std::ifstream prevEqn(prev_eqn.data());
    
    std::map<unsigned int, bool> hb_properties = prop_conv.get_hb_properties();
    std::map<unsigned int, bool> mm_properties = prop_conv.get_mm_properties();
    std::vector<std::string> line_split;
    
    int unary_HBs = 0;
    for(int i = 0 ; i < added_clauses.size() ; i++){
        line_split = split(added_clauses.at(i), ' ');
        if(line_split.size() == 3)
            unary_HBs++;
    }
    
    std::string line;

    std::getline(prevEqn, line); //getting the first line to get some information
    std::vector<std::string> first_line = split(std::string(line), ' ');
    std::string numOfVars = first_line.at(2);
    std::string numOfClauses = intToString(numOfOrigClauses);
    std::string weight = first_line.at(4);
    
    newEqn << "p wcnf " << numOfVars << " " << atoi(numOfClauses.c_str())+added_clauses.size()-unary_HBs+hb_properties.size() << " " << weight;   //writing the header line
    while (numOfOrigClauses > 0){
        std::getline(prevEqn, line);
        newEqn << "\n" << line;
        std::cout << "\n" << line;
        numOfOrigClauses--;
    }
    std::cout << std::endl;
    
    //Finally, adding the disjunction of negation of each of unsat cores lines:
    for(int i = 0 ; i < added_clauses.size() ; i++){
        line = added_clauses.at(i);
        line_split = split(line, ' ');
        if(line_split.size() != 3)
            newEqn << "\n" << line;
    }
    
    //then, writing the counterexample hb assignments part and the counterexample input arguments
    std::ifstream ceAssgnmnt(ce_assignment.data());
    while(std::getline(ceAssgnmnt, line)){
        line_split = split(line, ' ');

        if(line_split.at(0) == "s" && line_split.at(1) == "OPTIMUM" && line_split.at(2) == "FOUND")
            break;
    }
    
    std::cout << "size: " << input_arg_properties.size() << "\n";
    std::string word;
    bool print = true;
    int cur_var;
    while(ceAssgnmnt >> word){
        cur_var = atoi(word.c_str());
        if(word == "c")    //this means the hb variable assignments has finished and cursor moved to the next line
            break;
        else if(cur_var == hb_var_start || cur_var == -hb_var_start){
            print = true;
        }
        else if(std::find(input_arg_properties.begin(), input_arg_properties.end(), cur_var) != input_arg_properties.end()){
            newEqn << "\n" << weight << " " << cur_var << " 0";
        }
        else if(std::find(input_arg_properties.begin(), input_arg_properties.end(), -cur_var) != input_arg_properties.end()){
            newEqn << "\n" << weight << " " << cur_var << " 0";
        }
        
        if(print == true){
            literalt lit = literalt(abs(cur_var),false);
            
            std::string wgt = "1";   //default weight for CE variables is 1 (soft clause)

            if(hb_properties.count(lit.var_no()) > 0){
                //let's see if this variable belongs to one of the unary clauses in added_clauses
                std::string var_itself = weight + " " + intToString(cur_var) + " 0";
                std::string var_its_negation = weight + " " + intToString(-cur_var) + " 0";
                if(std::find(added_clauses.begin(), added_clauses.end(), var_itself) != added_clauses.end()){ //unit clauses (hbs)
                    newEqn << "\n" << var_itself;
		    std::cout << __FUNCTION__ << ":" << new_eqn << "(var_itself): " << wgt << " " << word << " 0" << std::endl; //added by chao
                }
                else if(std::find(added_clauses.begin(), added_clauses.end(), var_its_negation) != added_clauses.end()){
                    newEqn << "\n" << var_its_negation;
		    std::cout << __FUNCTION__ << ":" << new_eqn << "(var_its_negation): " << wgt << " " << word << " 0" << std::endl; //added by chao
                }
                else{     
	  	    if((hb_properties.find(lit.var_no()))->second)  { //if it is fixed (hbh) 
		        wgt = weight; 
                    } else {
                        wgt = "1";
		    }
                    newEqn << "\n" << wgt << " " << word << " 0";
		    std::cout << __FUNCTION__ << ":" << new_eqn << ": " << wgt << " " << word << " 0" << std::endl; //added by chao
                }
            }
	    else { //chao:: debugging only
	      if(mm_properties.count(lit.var_no()) > 0)
	      std::cout << __FUNCTION__ << ":" << new_eqn << ":  -mm-  " << word << std::endl; 
	    }
	}
    }
    
    newEqn.close();
    prevEqn.close();
    ceAssgnmnt.close();
}

/*******************************************************************\

Function: bmct::unsat_cores

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
void bmct::unsat_core(prop_convt &prop_conv){
    /*SEPI
     * 1- Feed unsat_eqn# file to compute_unsat_core function
     * 2- compute_unsat_core function feeds unsat_eqn# to MSUnCore and on each of its iterations
     * finds one or more clauses. Adds them as hard clauses to the file unsat_eqn## and goes on to
     * another iteration. At end, it reaches a state where the formula becomes unsatisfiable and 
     * no other unsat clause can be found. At that point, it has found a whole unsat core and returns.
     * 3- The return value of compute_unsat_core is the set of new clauses added to the cnf fed 
     * into it. We use this to write a new file check_eqn# (with non-flipped assertion) to see if 
     * we can find another counterexample. If yes, step 4. If not, step 5.
     * 4- Having the new counterexamle, we repeat what we did in write_to_file and at the end of 
     * build_goto_trace. And feed this new file again to compute_unsat_core.
     * 5- We have found all the unsat cores. There is no more counterexample. So, we return and 
     * finish the program.
     */
    /*SEPI: Now, I am passing the unsat_eqn file to msuncore to get the UNSAT core(s) until the non-flipped assertion formula is unsatisfiable*/
    std::vector<std::string> new_unsat_core;
    
    bool unsatisfiable = false; //if unsatisfiable is true, it means there is no (other) unsat core or counterexample left we might find
    int round = 0;
    
    while(!unsatisfiable){
        new_unsat_core = compute_unsat_core(iterNo++, round);
        if(new_unsat_core.size() == 0){
            std::cout << "THIS SHOULD NEVER HAPPEN, RIGHT?! PROBLEM!!" << std::endl;
            std::cout << "## iteration " << iterNo-1 << " UNSATISFIABLE!" << std::endl;
            //unsatisfiable = true;
            break;
        }
        
        std::cout << "## iteration " << iterNo-1 << " unsat core found is:" << std::endl;
        for(int i = 0 ; i < new_unsat_core.size() ; i++){
            std::cout << new_unsat_core.at(i) << std::endl;
        }
        
        // Now that we have an unsat core in hand, let's add it's negation to the disjunction-of-negation-of-assertions cnf formula:
        std::string new_eqn("/tmp/check_eqn" + intToString(iterNo) + ".cnf");
        
        std::string prev_eqn("/tmp/unsat_eqn" + intToString(iterNo-1) + /*"_" + intToString(round-1) +*/ ".cnf");
        
        write_check_file(prop_conv.get_clause_count() , new_eqn, prev_eqn, new_unsat_core, prop_conv.get_assertions());    //generates check_eqn#.cnf file
        
        if(do_check(new_eqn) == false){  //false means UNSATISFIABLE
            unsatisfiable = true;   // This means we can terminate the program. No more bug!
            std::cout << "Good news! Verification process finished!" << " (in " << iterNo << " iterations)\n";
        }
        else{
            std::cout << "Still not there!" << "\n";
            std::string new_unsat_eqn("/tmp/unsat_eqn" + intToString(iterNo) + ".cnf");
            std::string ce_assigment("/tmp/unsat_tmp");
            write_new_unsat_file(prop_conv, new_unsat_eqn, prev_eqn, ce_assigment, new_unsat_core);
        }
   
        std::cout << "------------------------------------" << std::endl;
    }
    
    
    std::cout << "\n##### The result of the verification is: (The extra constraints added on hb properties)" << std::endl;
    for(int i=0 ; i < added_clauses.size() ; i++){
        std::cout << added_clauses.at(i) << std::endl;
    }
    
    write_added_clauses_info();
}
/*******************************************************************\

Function: bmct::do_check

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
bool bmct::do_check(std::string eqn_file){   /*SEPI*/
    char cmd[9064];
    std::cout<<"--- Checking the effect of changes --- : "<<std::endl;

    //sprintf(cmd, "%s -u -v 0 %s > %s.out", "~/Desktop/co/src/cbmc/msuncore_20090525/bin/msuncore",eqn_file.data(),eqn_file.data());
    sprintf(cmd, "%s -u -v 0 %s > %s.out", MSUNCORE_LOC,eqn_file.data(),eqn_file.data());

    system(cmd);
    sprintf(cmd, "cp %s.out /tmp/unsat_tmp",eqn_file.data()); //chao, save check_eqn#.cnf.out for debugging
    system(cmd);
        
    std::ifstream unsatTmp("/tmp/unsat_tmp");
    
    std::string line;
    std::vector<std::string> line_split;
    while(std::getline(unsatTmp, line)){
        line_split = split(line, ' ');

        if(line_split.at(0) == "s" && line_split.at(1) == "UNSATISFIABLE"){
            line_split.clear();
            unsatTmp.close();
            return false;
        }
    }
    unsatTmp.close();
    return true;
}
/*******************************************************************\

Function: bmct::write_added_clauses_info

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
void bmct::write_added_clauses_info(){   /*SEPI*/
    std::cout << "\n##-------------------------------------------" << "\n";
    std::cout << "## Capturing the added clauses' information:" << "\n";
    std::ifstream hb_map("/tmp/hb_map");
    std::ifstream core_info("/tmp/core_info");
    
    std::string line;
    std::vector<std::string> line_split;
    
    std::map<std::string, std::string> varNo_varName;
    std::map<std::string, std::string> varName_varInfo;
    
    while(std::getline(hb_map, line)){
        line_split = split(line, ',');
        if(line_split.size() != 2){std::cout << "Error in reading file /tmp/hb_map\n"; return;}
        varNo_varName.insert(std::make_pair(line_split.at(0), line_split.at(1)));
    }
    while(std::getline(core_info, line)){
        line_split = split(line, ',');
        if(line_split.size() == 2)
            varName_varInfo.insert(std::make_pair(line_split.at(0), line_split.at(1)));
    }
    
    std::map<int,int> varNos;
    std::vector<std::string> vars;
    for(int i = 0; i < added_clauses.size() ; i++){
        vars = split(added_clauses.at(i), ' ');
        //to ignore the weight and the 0 at the end of the clause:
        for(int j = 1 ; j < vars.size()-1 ; j++){
            int varNo = abs(atoi(vars.at(j).c_str()));
	    if (varNos.find(varNo) == varNos.end())
	      varNos[varNo] = 1;
	}
    }
    for(std::map<int,int>::iterator vit = varNos.begin(); vit != varNos.end(); vit++) {
      int varNo = vit->first;
      std::string varName = (varNo_varName.find(intToString(varNo)))->second;
      std::string varInfo = (varName_varInfo.find(varName))->second;
      //std::cout << "\nVariable: " << varNo << " , Name: " << varName << "\n";
      //std::cout << "Means: " << varInfo << "\n";
      std::cout << varNo << ' ' << varInfo << '\n';
    }
    
    return;
}
/*******************************************************************\

Function: bmct::do_conversion

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::do_conversion(prop_convt &prop_conv)
{
  // convert HDL
  do_unwind_module(prop_conv);

  // convert SSA
  equation.convert(prop_conv);

  // the 'extra constraints'
  forall_expr_list(it, bmc_constraints)
    prop_conv.set_to_true(*it);
  
  /*SEPI: filling all_assertions in bmc.h with the assertions vector from prop.h now that it is ready */
  std::vector<int> v_assrt = prop_conv.get_assertions();
  for(int i=0 ; i < v_assrt.size() ; i++){
      all_assertions.push_back(intToString(v_assrt.at(i)));   //I needed the assertions themselves in bmc.cpp, not them in their violated form (in terms of +,-)
      std::cout << "Assertion: " << intToString(v_assrt.at(i)) << std::endl;
  }
  /*SEPI*/
}

/*******************************************************************\

Function: bmct::run_decision_procedure

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

decision_proceduret::resultt
bmct::run_decision_procedure(prop_convt &prop_conv)
{
   status() << "Passing problem to " 
            << prop_conv.decision_procedure_text() << eom;

  prop_conv.set_message_handler(get_message_handler());
  prop_conv.set_verbosity(get_verbosity());

  // stop the time
  absolute_timet sat_start=current_time();
  
  do_conversion(prop_conv);  

   status() << "Running " << prop_conv.decision_procedure_text() << eom;

   //std::ofstream myout("/tmp/bmc.cnf");
   //write_dimacs(myout);

  decision_proceduret::resultt dec_result=prop_conv.dec_solve();
  // output runtime

  {
    absolute_timet sat_stop=current_time();
     status() << "Runtime decision procedure: "
              << (sat_stop-sat_start) << "s" << eom;
  }

  return dec_result;
}

/*******************************************************************\

Function: bmct::report_success

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::report_success()
{
  result() << "VERIFICATION SUCCESSFUL" << eom;

  switch(ui)
  {
  case ui_message_handlert::PLAIN:
    break;
    
  case ui_message_handlert::XML_UI:
    {
      xmlt xml("cprover-status");
      xml.data="SUCCESS";
      std::cout << xml;
      std::cout << "\n";
    }
    break;
    
  default:
    assert(false);
  }
}

/*******************************************************************\

Function: bmct::report_failure

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::report_failure()
{
  result() << "\n" << "VERIFICATION FAILED" << eom;
  
  switch(ui)
  {
  case ui_message_handlert::PLAIN:
    break;
    
  case ui_message_handlert::XML_UI:
    {
      xmlt xml("cprover-status");
      xml.data="FAILURE";
      std::cout << xml;
      std::cout << "\n";
    }
    break;
    
  default:
    assert(false);
  }
}

/*******************************************************************\

Function: bmct::show_program

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::show_program()
{
  unsigned count=1;

  languagest languages(ns, new_ansi_c_language());
  
  std::cout << "\n" << "Program constraints:" << "\n";

  for(symex_target_equationt::SSA_stepst::const_iterator
      it=equation.SSA_steps.begin();
      it!=equation.SSA_steps.end(); it++)
  {
    if(it->is_assignment())
    {
      std::string string_value;
      languages.from_expr(it->cond_expr, string_value);
      std::cout << "(" << count << ") " << string_value << "\n";

      if(!it->guard.is_true())
      {
        languages.from_expr(it->guard, string_value);
        std::cout << std::string(i2string(count).size()+3, ' ');
        std::cout << "guard: " << string_value << "\n";
      }
      
      count++;
    }
    else if(it->is_assert())
    {
      std::string string_value;
      languages.from_expr(it->cond_expr, string_value);
      std::cout << "(" << count << ") ASSERT("
                << string_value <<") " << "\n";

      if(!it->guard.is_true())
      {
        languages.from_expr(it->guard, string_value);
        std::cout << std::string(i2string(count).size()+3, ' ');
        std::cout << "guard: " << string_value << "\n";
      }

      count++;
    }  
  }
}

/*******************************************************************\

Function: bmct::run

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool bmct::run(const goto_functionst &goto_functions)
{
  const std::string mm=options.get_option("mm");
  std::auto_ptr<memory_model_baset> memory_model(0);
  if(mm.empty() || mm=="sc")
    memory_model.reset(new memory_model_sct(ns));
  else if(mm=="tso")
    memory_model.reset(new memory_model_tsot(ns));
  else if(mm=="pso")
    memory_model.reset(new memory_model_psot(ns));
  else
    throw "Invalid memory model "+mm+" -- use one of sc, tso, pso";

  //symex.total_claims=0;
  symex.set_message_handler(get_message_handler());
  symex.set_verbosity(get_verbosity());
  symex.options=options;

   status() << "Starting Bounded Model Checking" << eom;

  symex.last_location.make_nil();

  try
  {
    // get unwinding info
    setup_unwind();

    // perform symbolic execution
    symex(goto_functions);

    // add a partial ordering, if required    
    if(equation.has_threads())
    {
      memory_model->set_message_handler(get_message_handler());
      (*memory_model)(equation);
    }
  }

  catch(std::string &error_str)
  {
    message_streamt message_stream(get_message_handler());
    message_stream.err_location(symex.last_location);
    message_stream.error(error_str);
    return true;
  }

  catch(const char *error_str)
  {
    message_streamt message_stream(get_message_handler());
    message_stream.err_location(symex.last_location);
    message_stream.error(error_str);
    return true;
  }

  catch(std::bad_alloc)
  {
    error() << "Out of memory" << eom;
    return true;
  }

  statistics() << "size of program expression: "
               << equation.SSA_steps.size()
               << " steps" << eom;

  try
  {
    if(options.get_option("slice-by-trace")!="")
    {
      symex_slice_by_tracet symex_slice_by_trace(ns);

      symex_slice_by_trace.slice_by_trace
	(options.get_option("slice-by-trace"), equation);
    }

    if(equation.has_threads())
    {
      // we should build a thread-aware SSA slicer
       statistics() << "no slicing due to threads" << eom;
    }
    else
    {
      if(options.get_bool_option("slice-formula"))
      {
        slice(equation);
        statistics() << "slicing removed "
                     << equation.count_ignored_SSA_steps()
                     << " assignments" << eom;
      }
      else
      {
        simple_slice(equation);
        statistics() << "simple slicing removed "
                     << equation.count_ignored_SSA_steps()
                     << " assignments" << eom;
      }
    }

    {
      statistics() << "Generated " << symex.total_claims
                   << " VCC(s), " << symex.remaining_claims
                   << " remaining after simplification" << eom;
    }

    if(options.get_bool_option("show-vcc"))
    {
      show_vcc();
      return false;
    }
    
    if(options.get_bool_option("cover-assertions"))
    {
      satcheckt satcheck;
      satcheck.set_message_handler(get_message_handler());
      satcheck.set_verbosity(get_verbosity());
      bv_cbmct bv_cbmc(ns, satcheck);
      bv_cbmc.set_message_handler(get_message_handler());
      bv_cbmc.set_verbosity(get_verbosity());

      if(options.get_option("arrays-uf")=="never")
        bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
      else if(options.get_option("arrays-uf")=="always")
        bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
    
      cover_assertions(goto_functions, bv_cbmc);
      return false;
    }

    // any properties to check at all?
    if(!options.get_bool_option("program-only") &&
       symex.remaining_claims==0)
    {
      report_success();
      return false;
    }

    if(options.get_bool_option("all-claims"))
    {
      satcheckt satcheck;
      satcheck.set_message_handler(get_message_handler());
      satcheck.set_verbosity(get_verbosity());
      bv_cbmct bv_cbmc(ns, satcheck);
      bv_cbmc.set_message_handler(get_message_handler());
      bv_cbmc.set_verbosity(get_verbosity());

      if(options.get_option("arrays-uf")=="never")
        bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
      else if(options.get_option("arrays-uf")=="always")
        bv_cbmc.unbounded_array=bv_cbmct::U_ALL;

      return all_claims(goto_functions, bv_cbmc);
    }
    
    if(options.get_bool_option("boolector"))
      return decide_boolector();
    else if(options.get_bool_option("mathsat"))
      return decide_mathsat();
    else if(options.get_bool_option("cvc"))
      return decide_cvc();
    else if(options.get_bool_option("dimacs"))
      return write_dimacs();
    else if(options.get_bool_option("opensmt"))
      return decide_opensmt();
    else if(options.get_bool_option("refine"))
      return decide_bv_refinement();
    else if(options.get_bool_option("aig"))
      return decide_aig();
    else if(options.get_bool_option("smt1"))
      // this is the 'default' smt1 solver
      return decide_smt1(smt1_dect::BOOLECTOR);
    else if(options.get_bool_option("smt2"))
      // this is the 'default' smt2 solver
      return decide_smt2(smt2_dect::MATHSAT);
    else if(options.get_bool_option("yices"))
      return decide_yices();
    else if(options.get_bool_option("z3"))
      return decide_z3();
    else
    {
      if(options.get_bool_option("program-only"))
      {
        show_program();
        return false;
      }

      return decide_default();
    }
  }

  catch(std::string &error_str)
  {
    error(error_str);
    return true;
  }

  catch(const char *error_str)
  {
    error(error_str);
    return true;
  }

  catch(std::bad_alloc)
  {
    error() << "Out of memory" << eom;
    return true;
  }
}

/*******************************************************************\

Function: bmct::decide

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool bmct::decide(prop_convt &prop_conv)
{
  if(options.get_bool_option("beautify-pbs") ||
     options.get_bool_option("beautify-greedy"))
    throw "sorry, this solver does not support beautification";

  prop_conv.set_message_handler(get_message_handler());
  prop_conv.set_verbosity(get_verbosity());
  
  bool result=true;

  switch(run_decision_procedure(prop_conv))
  {
  case decision_proceduret::D_UNSATISFIABLE:
    result=false;
    report_success();
    break;

  case decision_proceduret::D_SATISFIABLE:
    error_trace(prop_conv);
    report_failure();
    break;

  default:
    error() << "decision procedure failed" << eom;
  }

  return result;
}

/*******************************************************************\

Function: bmct::setup_unwind

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::setup_unwind()
{
  const std::string &set=options.get_option("unwindset");
  unsigned int length=set.length();

  for(unsigned int idx=0; idx<length; idx++)
  {
    std::string::size_type next=set.find(",", idx);
    std::string val=set.substr(idx, next-idx);

    long thread_nr=-1;
    if(!val.empty() &&
       isdigit(val[0]) &&
       val.find(":")!=std::string::npos)
    {
      std::string nr=val.substr(0, val.find(":"));
      thread_nr=atol(nr.c_str());
      val.erase(0, nr.size()+1);
    }

    if(val.rfind(":")!=std::string::npos)
    {
      std::string id=val.substr(0, val.rfind(":"));
      long uw=atol(val.substr(val.rfind(":")+1).c_str());

      symex.set_unwind_limit(id, thread_nr, uw);
    }
    
    if(next==std::string::npos) break;
    idx=next;
  }

  symex.max_unwind=options.get_int_option("unwind");
}
