/*******************************************************************\

Module:

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#ifndef CPROVER_PROP_H
#define CPROVER_PROP_H

// decision procedure wrapper for boolean propositional logics

#include <util/message.h>
#include <util/threeval.h>

#include "prop_assignment.h"

/*! \brief TO_BE_DOCUMENTED
*/
class propt:public messaget, public prop_assignmentt
{
public:
  /*SEPI*/
  propt() { this->assertion_flag = false; this->hb_var_start = 0; this->mm_var_start = 0;
  /*this->order_vars_start = 0;*/}
  /*SEPI*/
  virtual ~propt() {}
 
  /*SEPI*/
  unsigned int clause_cnt;
  unsigned int CE_var_start;
  unsigned int hb_var_start;
  //unsigned int order_vars_start;
  unsigned int mm_var_start;
  std::map<unsigned int, bool> hb_properties;   //if bool is true, it is a hbh constraint (hard clause)
  //std::map<unsigned int , bool> order_properties;
  std::map<unsigned int, bool> mm_properties;   //added by chao
  std::vector<literalt> VarLit;
  std::vector<std::string> SS;  //a stack to push all the clauses to. I will need this later
                                //to build the extended formula and write it into a file

  virtual int write_to_file(std::ostream &out, std::vector<int> violated_assertion, int args_no);
  //virtual void write_input_vars(std::ostream &out);
  /*SEPI*/
  
  // boolean operators
  virtual literalt land(literalt a, literalt b)=0;
  virtual literalt lor(literalt a, literalt b)=0;
  virtual literalt land(const bvt &bv)=0;
  virtual literalt lor(const bvt &bv)=0;
  virtual literalt lnot(literalt a)=0;
  virtual literalt lxor(literalt a, literalt b)=0;
  virtual literalt lxor(const bvt &bv)=0;
  virtual literalt lnand(literalt a, literalt b)=0;
  virtual literalt lnor(literalt a, literalt b)=0;
  virtual literalt lequal(literalt a, literalt b)=0;
  virtual literalt limplies(literalt a, literalt b)=0;
  virtual literalt lselect(literalt a, literalt b, literalt c)=0; // a?b:c
  virtual void set_equal(literalt a, literalt b);

  virtual void l_set_to(literalt a, bool value)
  {
    set_equal(a, const_literal(value));
  }
  
  void l_set_to_true(literalt a)
  { l_set_to(a, true); }
  void l_set_to_false(literalt a)
  { l_set_to(a, false); }

  // constraints
  virtual void lcnf(const bvt &bv)=0;
  virtual bool has_set_to() const { return true; }
  
  // assumptions
  virtual void set_assumptions(const bvt &_assumptions) { }
  virtual bool has_set_assumptions() const { return false; }

  // variables
  virtual literalt new_variable()=0;
  virtual void set_variable_name(literalt a, const std::string &name) { }
  virtual unsigned no_variables() const=0;
  bvt new_variables(unsigned width);
  
  // solving
  virtual const std::string solver_text()=0;
  typedef enum { P_SATISFIABLE, P_UNSATISFIABLE, P_ERROR } resultt;
  virtual resultt prop_solve()=0;  
  
  // satisfying assignment, from prop_assignmentt
  virtual tvt l_get(literalt a) const=0;
  virtual void set_assignment(literalt a, bool value);
  virtual void copy_assignment_from(const propt &prop);

  // returns true if an assumption is in the final conflict
  virtual bool is_in_conflict(literalt l) const;  
  virtual bool has_is_in_conflict() const { return false; }
  
  // an incremental solver may remove any variables that aren't frozen
  virtual void set_frozen(literalt a) { }

  /*SEPI*/
  void set_assertion_flag(bool b){assertion_flag=b;}
  bool is_assertion() {return assertion_flag;}
  void set_assertions(int var_no){assertions.push_back(var_no);}
  std::vector<int> get_assertions(){return assertions;}
  int get_clause_cnt(){return (int)(clause_cnt);}
  void set_soft(bool b){soft=b;}

  protected:
  bool assertion_flag;
  bool soft;
  std::vector<int> assertions;
  std::vector<literalt> po_constraitns;
  /*SEPI*/
};

#endif
