/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Implementation of an atomicity violation found in Apache 2.0.48, bug report #21287
 *
 * Source: http://web.eecs.umich.edu/~jieyu/bugs/apache-21287.html
 *
 * Bug Report: http://issues.apache.org/bugzilla/show_bug.cgi?id=21287
 *
 * Apache Source: http://web.eecs.umich.edu/~jieyu/bugs/apache-21287/httpd-2.0.48.tar.gz
 *
 * Note: A copy of the apache source is hosted on the git repo and I (markus)
 * should have a local copy as well.
 */
#include <pthread.h>
#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>

/* Comment out to remove debugging printf() calls and calls to sleep() to
 * attempt to force the buggy interleaving */
//#define DEBUG

#define INT_TYPE int8_t 

INT_TYPE refCount;

// flags to indicate the operations performed by the threads
bool t1_free;
bool t2_free;

void *decRefCount_t1(void *arg) {
  /* decrement the reference count */
  refCount--;

  /* if the reference count == 0, then free it */
  if (refCount == 0) t1_free = true;

  return NULL;
}

void *decRefCount_t2(void *arg) {
  /* decrement the reference count */
  refCount--;

  /* if the reference count == 0, then free it */
  if (refCount == 0) t2_free = true;

  return NULL;
}

/**
 * Initialze the global struct object. Spawn two threads to go out and
 * decrement the reference count then exit the main thread
 */
int main() {
  pthread_t t1, t2;

  // Initialize everything
  refCount = 2;	
  t1_free = false;
  t2_free = false;

  pthread_create(&t1, NULL, decRefCount_t1, NULL);
  pthread_create(&t2, NULL, decRefCount_t2, NULL);

  pthread_join(t1, NULL);
  pthread_join(t2, NULL);

  if (t1_free && t2_free) {
    assert(0);
  }

}
