/**
 * \file freeBSD_audit1.c
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Example program of a data race found in FreeBSD file audit_arg.c
 *
 * Source: Z. Yin, D. Yuan, Y. Zhou, et al. How Do Fixes Become Bugs? A
 * Comprehensive Study on Incorrect Fixes in Commercial and Open Source
 * Operating Systems. September 2011.
 *
 * This program demonstrates the original bug in audit_arg.c. There is a data
 * race on the variable unprotected in the function doWork(). 
 *
 * The code was greatly simplified, but the bug is hopefully still reflective. 
 */

#include <stdlib.h>
#include <pthread.h>
#include <stdint.h>
#include <assert.h>

/* Comment out to remove debugging messages */
//#define DEBUG

#ifdef DEBUG
#include <stdio.h>
#endif

#define INT_TYPE int8_t

//INT_TYPE unprotected;
int8_t unprotected;

void *doWork_t1() {
#ifdef DEUBG
    printf("Work Function Entered\n");
#endif 

    if (unprotected > 2) {
      return NULL;
    }

    unprotected++;

#ifdef DEBUG
    printf("The Value of unprotected is: %d\n", unprotected);
#endif

    return NULL;
}

void *doWork_t2() {
#ifdef DEUBG
    printf("Work Function Entered\n");
#endif 

    if (unprotected > 2) {
      	return NULL;
    }

    unprotected++;

#ifdef DEBUG
    printf("The Value of unprotected is: %d\n", unprotected);
#endif

    return NULL;
}

void *doWork_t3() {

    if (unprotected > 2) {
      	return NULL;
    }

    unprotected++;

#ifdef DEBUG
    printf("The Value of unprotected is: %d\n", unprotected);
#endif

    return NULL;
}

/**
 * Spawns three threads in the doWork() function 
 *
 * \return Returns NULL upon completion
 */
int main() {
    pthread_t t1, t2, t3;

    pthread_create(&t1, NULL, doWork_t1, NULL);
    pthread_create(&t2, NULL, doWork_t2, NULL);
    pthread_create(&t3, NULL, doWork_t3, NULL);

    //pthread_join(t1, NULL);
    //pthread_join(t2, NULL);
    //pthread_join(t3, NULL);

    if (unprotected != 2) {
      assert(0);
    }
}
