/*
 * Author: Markus Kusano
 *
 * Source: http://gcc.gnu.org/bugzilla/show_bug.cgi?id=21334
 */
#include <stdbool.h>
#include <stdint.h>
#include <pthread.h>
#include <assert.h>

// Data associated with the string "object"
char str_data;
// Lock protecting the string
pthread_mutex_t str_lock;
// True if the string is shared across multiple threads
bool str_is_shared;
// Number of references of the string being used
int8_t ref_count;
bool destroyed;


void *t1_main(void *args);

void *t1_main(void *args) {
  // Copy the string. This is simply used as an operation incrementing the
  // reference count
  if (str_is_shared) {
    pthread_mutex_lock(&str_lock);
    ref_count++;
    pthread_mutex_unlock(&str_lock);
  }

  // Call the string destructor
  pthread_mutex_lock(&str_lock);
  if (ref_count < 0) {
    // no more reference to the string so we can destroy it
    pthread_mutex_unlock(&str_lock);
    destroyed = true;
  }
  else {
    // still references out there, do nothing
    pthread_mutex_unlock(&str_lock);
  }
  // End destructor

  return NULL;
}

void *t2_main(void *args) {
  // Code from obtaining the .end() iterator. This function has strange
  // behavior which causes the bug
  //
  // Note: str_is_shared is always false so this branch is always taken
  if (!str_is_shared) {
    pthread_mutex_lock(&str_lock);
    // Set ref_count to -1. I think the intetion was for this to be an
    // optimization for the non-shared case.
    ref_count = -1;
    pthread_mutex_unlock(&str_lock);
  }

  // At this point, thread one can come in and destroy the mutex since
  // str_is_shared is false (so thread one will not increment the reference
  // count). The buggy behavior is that thread 2 can perform operations on the
  // thread after thread one destroys the string.
  
  // perform a modification
  if (destroyed) {
    assert(0);
  }
  str_data = 'a';

  return NULL;
}

int main() {
  pthread_t t1, t2;

  // Initialize everything
  pthread_mutex_init(&str_lock, NULL);
  str_is_shared = false;
  str_data = 'x';
  ref_count = 0;
  destroyed = false;

  pthread_create(&t1, NULL, t1_main, NULL);
  pthread_create(&t2, NULL, t2_main, NULL);

  pthread_join(t1, NULL);
  pthread_join(t2, NULL);

  return 0;

}
