/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Source: http://gcc.gnu.org/bugzilla/show_bug.cgi?id=40518
 *
 * Data race in GCC's libstdc++ in their implementation of std::string.erase()
 *
 * See README.md for more information
 */

#include <assert.h>
#include <pthread.h>

/**
 * The Rep object in std::string
 *
 * This can be viewed as a global Rep object held in a global string which is
 * modified by the threads in t1_main() and t2_main()
 */
int length;	///< The length of the string
int capacity;   ///< Max length of string
int ref_count;   ///< The number of references to the string

/**
 * Sets all the members of the global "rep object" to zero.
 *
 * According to the bug report, even if two threads are writing 0 this should
 * not happen. The fix is simply to bail out earlier (i.e., bail out if the
 * assertion is true).
 */
void erase();

/**
 * Calls erase() on the global Rep empty_rep
 *
 * \param args Unused
 * \return Returns NULL upon completion
 */
void *t1_main(void *args);
void *t2_main(void *args);

void erase() {
  if (length == 0) {
    assert(0);
  }
  length = 0;
  //if (capacity == 0) {
    //assert(0);
  //}
  capacity = 0;
  //if (ref_count == 0) {
    //assert(0);
  //}
  ref_count = 0;
}

void *t1_main(void *args) {
  // Bail out early
  if (length == 0) {
    return NULL;
  }
  // Otherwise, erase. There is still a data race!
  // erase();
  int tmp = length;
  if (tmp == 0) {
    assert(0);
  }
  length = 0;
  capacity = 0;
  ref_count = 0;

  return NULL;
}

void *t2_main(void *args) {
  // Bail out early
  if (length == 0) {
    return NULL;
  }
  // Otherwise, erase. There is still a data race!
  // erase();
  int tmp = length;
  if (tmp == 0) {
    assert(0);
  }
  length = 0;
  capacity = 0;
  ref_count = 0;

  return NULL;
}

/**
 * Initializes the global Rep empty_rep and creates threads in t1_main() and
 * t2_main()
 *
 * \return Returns 0 upon completion
 */
int main(void) {
    pthread_t t1, t2;

    length = 1;
    capacity = 1;
    ref_count = 1;

    pthread_create(&t1, NULL, t1_main, NULL);
    pthread_create(&t2, NULL, t2_main, NULL);

    pthread_join(t1, NULL);
    //pthread_join(t2, NULL);

    return 0;
}
