/**
 * \file glib_data_race.c
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Source: https://bugzilla.gnome.org/show_bug.cgi?id=512624
 *
 * Data race and atomicity violoation found in glib.
 *
 * Involves the use to unatomic increment/decrements in multiple threads
 */

#include <pthread.h>
#include <assert.h>

/**
 * Value that is incremented or decremented by t1_main() or t2_main()
 */
int shared;

/**
 * Increments shared a bunch
 *
 * \param args Unused
 * \return Returns NULL upon completion
 */
void *t1_main(void *args);

/**
 * Decrements shared a bunch
 *
 * \param args Unused
 * \return Returns NULL upon completin
 */
void *t2_main(void *args);

void *t1_main(void *args) {
    
	  shared++;
	  shared++;
	  //shared++;
	  //shared++;
	  //shared++;
	  //shared++;
	  //shared++;
	  //shared++;
	  //shared++;
	  //shared++;

    return NULL;
}

void *t2_main(void *args) {

	  shared--;
	  shared--;
	  //shared--;
	  //shared--;
	  //shared--;
	  //shared--;
	  //shared--;
	  //shared--;
	  //shared--;
	  //shared--;

    return NULL;
}

/**
 * Initialized shared to 0 and then spawns threads in t1_main() and t2_main.
 * calls pthread_join on both the threads and checks if the end value of shared
 * is 0 (using assert()).
 *
 * \param argc Unused
 * \param argv Unused
 * \return Returns NULL upon completion
 */

int main(int argc, char *argv[]) {
    /* threads for t1_main and t2_main */
    pthread_t t1, t2;

    /* initialize shared */
    shared = 0;

    /* create the threads */
    pthread_create(&t1, NULL, t1_main, NULL);
    pthread_create(&t2, NULL, t2_main, NULL);

    /* wait for them to exit */
    pthread_join(t1, NULL);
    pthread_join(t2, NULL);

    /* the two thread should decrement and increment the same number of times */
    assert(shared == 0);

    return 0;
}

