// Author: Markus Kusano
//
// A sequential array based list. 
//
// The test program creates two thread which adds to the list, joins the
// threads, and then checks that the final state of the list is consistent with
// the operations performed by the threads.
#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>
#include <assert.h>
#include <pthread.h>

//The maxmimum number of the hash table lists
#define BUCKETS 2

// The maximum size of the array representing the list. The entire array is
// allocated at the start
#define MAX_SIZE 2

// Type of data stored in the list
#define NODE_TYPE char

// Type of the variable representing the next open slot in the list
#define SIZE_TYPE uint8_t

typedef struct list {
  // The array of items in the list
  NODE_TYPE arr[MAX_SIZE];
  // integer of the next open location in the list
  SIZE_TYPE open;
} list_t;

typedef struct __hash_t {
//  list_t gl[BUCKETS]
  list_t lists[BUCKETS];
} hash_t;

// Global hash table. For convience so we do not need to deal with passing parameters
// to the threads
hash_t gh;

// Add an item with value `i` to the passed list 
void list_add(list_t *s, NODE_TYPE i);

// Remove the item with the value `i` from the list. Returns false if the item
// was not remove (i.e., not found). Otherwise, returns true.
bool list_remove(list_t *s, NODE_TYPE i);

// Returns true if the passed list contains the passed item
bool list_contains(list_t *s, NODE_TYPE i) {
  // Manually unrolled
  /*
  for (SIZE_TYPE j = 0; j < MAX_SIZE; ++j) {
    SIZE_TYPE cur = s->arr[j];
    if (cur == i) {
      return true;
    }
  }
  */

  /* Begin for loop */
  int j = 0;
  /* ========== j == 0 ========== */
  SIZE_TYPE cur = s->arr[j];
  if (cur == i) {
    return true;
  }
  /* ========== j == 0 ========== */
  /* ========== j == 1 ========== */
  j += 1;
  SIZE_TYPE cur = s->arr[j];
  if (cur == i) {
    return true;
  }
  /* ========== j == 1 ========== */
  /* End for-loop */
  return false;
}

bool hash_lookup(hash_t *H, int key) {
  int bucket = key % BUCKETS;
  return list_contains(&H->lists[bucket], key);
}

void list_add_t1(list_t *s, NODE_TYPE i) {
  // Insert the item in to the array
  s->arr[s->open] = i;

  // Update the next open slot of the array
  s->open += 1;
}

void hash_insert_t1(hash_t *H, NODE_TYPE key) {
  int bucket = key % BUCKETS;
  list_add_t1(&H->lists[bucket], key);
}

void list_add_t2(list_t *s, NODE_TYPE i) {
  // Insert the item in to the array
  s->arr[s->open] = i;

  // Update the next open slot of the array
  s->open += 1;
}

void hash_insert_t2(hash_t *H, NODE_TYPE key) {
  int bucket = key % BUCKETS;
  list_add_t2(&H->lists[bucket], key);
}

// initialize the passed list
void list_init(list_t *s) {
  s->open = 0;
}

// initialize the passed hash table
void hash_init(hash_t *H) {
  int i;
  for(i=0; i < BUCKETS; i++) {
	  list_init(&H->lists[i]);
  }
}

void *t1_main(void *args) 
{
  hash_insert_t1(&gh, 1);
  return NULL;
}

void *t2_main(void *args) 
{
  hash_insert_t2(&gh, 2);
  return NULL;
}

int main() {
  hash_init(&gh);

  pthread_t t1;
  pthread_t t2;

  pthread_create(&t1, NULL, t1_main, NULL);
  pthread_create(&t2, NULL, t2_main, NULL);

  pthread_join(t1, NULL);
  pthread_join(t2, NULL);

  if(!hash_lookup(&gh, 1)) {
	assert(0);
  }
 /* if (!hash_lookup(&gh, 2)) {
    assert(0);
  }*/
  return 0;
}
