/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Source: S. Lu, S. Park, E. Seo, Y. Zhou. Learning from Mistakes - A
 * Comprehensive Study on Real World Concurrency Bug Characteristics
 *
 * Bug Report: https://bugzil.la/61369
 *
 * Example of an atomicity violation and data race found in mozilla.
 *
 * See README.md for more information.
 */

#include <pthread.h>
#include <stdlib.h>

/* Comment out to disable debugging information */
//#define DEBUG

/**
 * Struct that in the mozilla code holds more information, but the variable
 * this program is concerned with is only state, which is what the bug is
 * circled around
 */
struct mthread_ {
    char state;    /**< Variable that is raced after */
} mthread;  /**< instance of the mthread_ struct */

/**
 * Initialize the thread structure.
 *
 * \param arg Unused
 * \return Returns NULL upon completion
 */
void *threadInit(void *arg);

void *threadInit(void *arg) {
    mthread.state = 'A'; // A for allocated!

    return NULL;
}

/** 
 * Spawn a thread to initialize the thread structure and then access the state
 * information. There is assumed atomicity between these operations but no
 * synchronization exisits
 *
 * \return Returns 0 upon completion
 */
int main() {
    pthread_t initer;

    char getState;

    mthread.state = 'X'; // X for BAD!

    pthread_create(&initer, NULL, threadInit, NULL);

    // Access the state information, the atomicity violation occurs here. It is
    // assumed that the state has been initialized but no synchornization
    // variables are in place to ensure that.

    getState = mthread.state;
    assert(getState != 'X');

    pthread_exit((void *) 0);
}
