/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 * Simplified implementation of a MySQL atomicity and data race violation.
 *
 * Source: S. Lu, S. Park, E. Seo, Y. Zhou. Learning Mistakes - A
 * Comprehensive
 * Study on Real World Concurrency Bug Characteristics
 *
 * MySQL Source:
 * http://web.eecs.umich.edu/~jieyu/bugs/mysql-3596/mysql-4.0.19.tar.gz
 *
 * See README.md for more details
 */
#include <pthread.h>
#include <stdlib.h>

/* Commment out the following to disable debugging information */
//#define DEBUG

#ifdef DEBUG
#include <stdio.h>
#endif

/**
 * Data that is raced after 
 */
int *sharedData;

/**
 * Calls free() and sharedData()
 */
void *delete();

/**
 * Increments sharedData. This accomplishes using the data to potentialy cause
 * an atomicity violation 
 */
void *use();

void *delete() {
    free(sharedData);
    sharedData = NULL;

    #ifdef DEBUG
    printf("Data freed\n");
    #endif

    return NULL;
}

void *use() {
    #ifdef DEBUG
    printf("About to use the data\n");
    #endif
    assert(sharedData != NULL);

    return NULL;
}

/**
 * Allocates memory for sharedData and creates two threads: one in use() and
 * one in delete()
 *
 * \return Returns 0 upon completion
 */
int main() {
    /* Allocate memory for the shared data */
    sharedData = malloc(sizeof(int));

    pthread_t user, deleter;

    pthread_create(&user, NULL, use, NULL);
    pthread_create(&deleter, NULL, delete, NULL);

    /* Prevent main() from "falling off" and terminating the process before the
     * threads complete */

    pthread_exit(NULL);
}



