/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Source: http://web.eecs.umich.edu/~jieyu/bugs/mysql-644.html
 *
 * Much simplified version of a bug found in MySQL. See README.md for more
 * information.
 */
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* Comment out to remove debugging calls to printf() */
//#define DEBUG

/* Comment out to remove calls to sleep() with attempt to force the bug */
//#define FORCE

#ifdef FORCE
#include <unistd.h>
#endif

/**
 * Table that is contended for. The table is as a character representing it
 * being valid or invalid
 */
char table1 = 'A'; // A for allocated!

/**
 * Another table that is contended for. It is the same as table1. These two
 * tables are joined in `join_init_cache()`.
 */
char table2 = 'A'; // A for allocated!

/**
 * Iterates over the tables and checks if the tables have equal fields
 *
 * \param args Unused
 * \return Return's NULL upon completion
 */
void *join_init_cache(void *args);


/**
 * Removes a two fields from the end of table2
 *
 * \param args Unused
 * \return Return's NULL upon completion
 */
void *fix_field(void *args);

void *join_init_cache(void *args) {
#if 0
    unsigned table_1_size;
    unsigned table_2_size;

    /* Number of fields in the tables that are the same */
    int the_same;

    table_1_size = 0;
    table_2_size = 0;
    the_same = 0;

    /* Pointer to examine the entries in the tables */
    char *cur;

    /* Calculate the size of table1 */
    for (cur = table1; *cur != '\0'; cur++) {
	    table_1_size++;
    }

    /* Calculate the size of table2 */
    for (cur = table2; *cur != '\0'; cur++) {
	    table_2_size++;
    }

    /* Compare all the fields to each other. This really does nothing but when
     * combined with fix_field() running in another thread it could result in a
     * segfault */
    int i, j;
    for(i = 0; i < table_1_size; i++) {
	    for (j = 0; j < table_2_size; j++) {
        assert(table2 != NULL);
	      if (table1[i] == table2[j]) {
		      the_same++;
	    }
	  }
    }
#endif

    // Check that the state of both tables is valid
    assert(table1 != 'X');
    assert(table2 != 'X');

    return NULL;
}

void *fix_field(void *args) {
    
#if 0
    /* temporary string to hold the modifications to table2 */
    char *temp;
    /* The size of table 2 */
    //unsigned size;

    /* Size starts at one because the for loop below does not include the null
     * terminator in the size */
    size = 1;

    /* Calcuate the size of table 2 */
    for (temp = table2; *temp != '\0'; temp++) {
	    /ize++;
    }

    /* Remove two characters from the table so: */
    size -= 2;

    /* copy over all but the last two characters from table2 */
    temp = malloc(sizeof(char) * size);
    memcpy(temp, table2, size - 1);

    /* Place in the null terminator */
    memcpy(temp + size, "\0", 1);

    /* free the memory for table2 to avoid leaks */
    free(table2);
    table2 = NULL;

    table2 = temp;
#endif

    char tmp = table2;
    tmp = 'B'; // B for Best!
    table2 = 'X'; // X for unallocated :C
    table2 = tmp;

    return NULL;
}

/**
 * Initializes the tables and then spawns threads in join_init_cache() and
 * fix_field()
 *
 * \return Returns 0 upon completion
 */
int main(void) {
    pthread_t t1, t2;

    /* allocate memory for the tables */
    //table1 = malloc(sizeof(char) * 8);
    //table2 = malloc(sizeof(char) * 8);

    /* Initialize the tables */
    //memcpy(table1, "abcdefg\0", 8);
    //memcpy(table2, "hijklmn\0", 8);

    pthread_create(&t1, NULL, join_init_cache, NULL);
    pthread_create(&t2, NULL, fix_field, NULL);

    pthread_join(t1, NULL);
    pthread_join(t2, NULL);
}

