#!/bin/bash

# Test running script: Calls CBMC and friends. Assumes all the environment
# variables are setup (see exports.sh)

set -u

#source exports.sh

die () {
  if [ ! -z "${@:-}" ]
  then
    echo ${@}
  fi
  exit 1
}

if [ -z ${1:-} ]
then
  die "First argument must be program to test"
fi

if [ ! -f ${1:-} ]
then
  die "Program not found: $1"
fi

if [ ! -f $CBMC ]
then
  die "CBMC binary not found: $CBMC"
fi

if [ ! -f $MAKER ]
then
  die "EdgeMaker binary not found: $MAKER"
fi

if [ ! -f $FIXER ]
then
  die "EdgeFixer binary not found: $FIXER"
fi

if [ ! -f $PREPROC ]
then
  die "Pre-processor script not found: $PREPROC"
fi

# Temporary files for various programs
#CBMC_OUT=/tmp/cbmc.out
#FIX_OUT=/tmp/fixer.out

CBMC_OUT=./cbmc.out
FIX_OUT=./fixer.out
TIME=`which time`

# cleanup old files?
rm -f /tmp/*.out

# First, run CBMC
#>&2 echo "Running CBMC..."
#{ time $CBMC $1 >$CBMC_OUT 2>/dev/null ; } 2>./cbmc.time
{ $TIME -f '%e' $CBMC $1 >$CBMC_OUT ; } 2>cbmc.time
tail -n 1 cbmc.time > /tmp/cbmc.time
mv /tmp/cbmc.time cbmc.time

#>&2 echo "Running Preprocessor..."
{ $TIME -f '%e' $PREPROC ${CBMC_OUT} ; } 2>pp.time

#>&2 echo "Running Edge Maker..."
{ $TIME -f '%e' $MAKER >$FIX_OUT ; } 2>make.time

#>&2 echo "Running Program Repair..."
{ $TIME -f '%e' $FIXER $1 $FIX_OUT ; } 2>rep.time

echo "`cat pp.time` + `cat make.time` + `cat rep.time`" | bc >repair.time
