#!/bin/bash

# Given a directory (as argument 1) test it
# Assumes exports are already setup (see exports.sh)

# =============== Variable Definitions ==================
# Name of C files in each test directory
MAIN="main.c"
# Name of expected output
EXPECTED="expected"

# Display an error message (if passed as argument 1) and exit with error code 1
die () {
  if [ -z "${1-}" ]
  then
    exit 1
  fi
  echo "Error: $1"
  exit 1
}

# Issue a warning and then exit with 0
warn () {
  if [ ! -z "${1-}" ]
  then
    echo "Warning: $1"
  fi
  exit 0
}

if [ -z "${1-}" ]
then
  die "Pass directory to test"
fi

if [ ! -e "${1-}" ]
then
  die "Test location does not exist ($1)"
fi

if [ ! -d "${1-}" ]
then
  die "Test location is not a directory ($1)"
fi

# CFILE=${1}/${MAIN}
# EXP=${1}/${EXPECTED}

source exports.sh
cd $1
CFILE=${MAIN}
EXP=${EXPECTED}


if [ ! -f "$CFILE" ]
then
  die "Test file not found ($CFILE)"
fi

# Perform test and check results
echo "Testing: $1"
../run.sh $CFILE >test.out
diff test.out $EXP || warn "${1}: tests dont match"

echo "     CBMC Time: `cat cbmc.time`"
echo "     Repair Time: `cat repair.time`"
echo "     `grep 'Num Solutions' fixer.out`"
