/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Example program documenting a data race/atomicity violation in 
 * Transmission-1.42.
 *
 * Source: http://web.eecs.umich.edu/~jieyu/bugs/transmission-142.html
 *
 * Transmission Source:
 * http://web.eecs.umich.edu/~jieyu/bugs/transmission-1.42/transmission-1.42.tar.bz2
 *
 * See README.md for more information
 */

#include <pthread.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>


/* Comment out to remove debugging messages */
//#define DEBUG

#ifdef DEBUG
#include <stdio.h>
#endif

/**
 * Represents a variable in transmission that is raced after 
 */
char *bandwidth;

/**
 * Thread function that mallocs bandwidth
 *
 * \param arg Unused
 * \return Returns NULL upon completion
 */
void *init(void *arg);

/**
 * Writes data to bandwidth. There is no atomicity between this and init()
 * which is where the bug comes from
 *
 * \param arg Unused
 * \return Returns NULL upon completion
 */
void *allocate(void *arg);

void *init(void *arg) {
    bandwidth = malloc(sizeof(char));
    return NULL;
}


void *allocate(void *arg) {
  if (bandwidth == NULL) { assert(0); }
  memcpy(bandwidth, "9", 1);
  return NULL;
}


/**
 * Sets bandwidht to NULL and spawns the thread functions
 *
 * \return Returns 0 upon completion
 */
int main() {
    pthread_t t1, t2;

    bandwidth = NULL;

    pthread_create(&t1, NULL, init, NULL);
    pthread_create(&t2, NULL, allocate, NULL);

    pthread_exit((void *) 0);
}
